from typing import Any, Optional, Callable, cast
from fiatlight import FunctionWithGui
from fiatlight.computer_vision import ImageUInt8
from fiatlight.computer_vision.image_with_gui import ImageWithGui, ImageChannelsWithGui
from fiatlight.computer_vision.cv_color_type import ColorConversion

import cv2
import numpy as np


def split_channels(image: ImageUInt8) -> ImageUInt8:
    assert len(image.shape) == 3
    depth_first = np.squeeze(np.dsplit(image, image.shape[-1]))
    return depth_first


class SplitChannelsWithGui(FunctionWithGui):
    color_conversion: Optional[ColorConversion] = None
    gui_params_optional_fn: Callable[[], bool]

    def __init__(self) -> None:
        self.input_gui = ImageWithGui()
        self.output_gui = ImageChannelsWithGui()

    def output_gui_channels(self) -> ImageChannelsWithGui:
        return cast(ImageChannelsWithGui, self.output_gui)

    def f(self, x: Any) -> Any:
        assert type(x) == np.ndarray
        if self.color_conversion is not None:
            x_converted = cv2.cvtColor(x, self.color_conversion.conversion_code)
        else:
            x_converted = x
        channels = split_channels(x_converted)  # type: ignore
        channels_normalized = channels / 255.0
        return channels_normalized

    def name(self) -> str:
        r = "Split Channels"
        if self.color_conversion is not None:
            r += " - " + self.color_conversion.name
        return r

    def gui_params(self) -> bool:
        if hasattr(self, "gui_params_optional_fn"):
            return self.gui_params_optional_fn()
        else:
            return False


class MergeChannelsWithGui(FunctionWithGui):
    color_conversion: Optional[ColorConversion] = None

    def __init__(self) -> None:
        self.input_gui = ImageChannelsWithGui()
        self.output_gui = ImageWithGui()

    def f(self, x: Any) -> Any:
        assert type(x) == np.ndarray
        channels = [c for c in x]
        image_float = np.dstack(channels)
        image_uint8 = (image_float * 255.0).astype("uint8")
        image_converted = image_uint8
        if self.color_conversion is not None:
            image_converted = cv2.cvtColor(image_uint8, self.color_conversion.conversion_code)
        return image_converted

    def name(self) -> str:
        r = "Merge Channels"
        if self.color_conversion is not None:
            r += " - " + self.color_conversion.name
        return r



from typing import List, Tuple
from imgui_bundle import immapp
from fiatlight.versatile import VersatileFunctionsCompositionGraph


poem = """
 If—
By Rudyard Kipling

If you can keep your head when all about you
    Are losing theirs and blaming it on you,
If you can trust yourself when all men doubt you,
    But make allowance for their doubting too;
If you can wait and not be tired by waiting,
    Or being lied about, don’t deal in lies,
Or being hated, don’t give way to hating,
    And yet don’t look too good, nor talk too wise:

If you can dream—and not make dreams your master;
    If you can think—and not make thoughts your aim;
If you can meet with Triumph and Disaster
    And treat those two impostors just the same;
If you can bear to hear the truth you’ve spoken
    Twisted by knaves to make a trap for fools,
Or watch the things you gave your life to, broken,
    And stoop and build ’em up with worn-out tools:

If you can make one heap of all your winnings
    And risk it on one turn of pitch-and-toss,
And lose, and start again at your beginnings
    And never breathe a word about your loss;
If you can force your heart and nerve and sinew
    To serve your turn long after they are gone,
And so hold on when there is nothing in you
    Except the Will which says to them: ‘Hold on!’

If you can talk with crowds and keep your virtue,
    Or walk with Kings—nor lose the common touch,
If neither foes nor loving friends can hurt you,
    If all men count with you, but none too much;
If you can fill the unforgiving minute
    With sixty seconds’ worth of distance run,
Yours is the Earth and everything that’s in it,
    And—which is more—you’ll be a Man, my son!
"""


WordWithCount = Tuple[str, int]


def str_lower(s: str) -> str:
    return s.lower()


def remove_non_letters(s: str) -> str:
    r = ""
    for c in s:
        if c.isalpha() or c == " ":
            r += c
        elif c.isspace():
            r += " "
    return r


def split_words(s: str) -> List[str]:
    r = s.split(" ")
    return r


def remove_empty_words(words: List[str]) -> List[str]:
    r = [word for word in words if len(word) > 0]
    return r


def run_length_encode(input_list: List[str]) -> List[WordWithCount]:
    r: List[WordWithCount] = []

    for i in range(len(input_list)):
        current = input_list[i]
        previous = input_list[i - 1] if i > 0 else None
        if current == previous:
            previous_count = r[-1][1]
            r[-1] = (current, previous_count + 1)
        else:
            r.append((current, 1))
    return r


def sort_word_with_counts(words: List[WordWithCount]) -> List[WordWithCount]:
    r = sorted(words, key=lambda w: w[1], reverse=True)
    return r


def display_word_with_counts(words: List[WordWithCount]) -> str:
    strs = [w[0] + ": " + str(w[1]) for w in words]
    r = "\n".join(strs)
    return r


def main() -> None:
    functions = [
        remove_non_letters,
        str_lower,
        split_words,
        remove_empty_words,
        sorted,
        run_length_encode,
        sort_word_with_counts,
        display_word_with_counts,
    ]
    functions_graph = VersatileFunctionsCompositionGraph(functions)  # type: ignore

    functions_graph.set_input(poem)

    def gui() -> None:
        functions_graph.draw()

    immapp.run(gui, with_node_editor=True, window_size=(1400, 600), window_title="Compose")


if __name__ == "__main__":
    main()



from __future__ import annotations
from typing import Any
from typing import Optional

from fiatlight.any_data_with_gui import AnyDataWithGui
from imgui_bundle import imgui


class IntWithGui(AnyDataWithGui):
    value: int = 0

    def gui_data(self, function_name: str) -> None:
        imgui.text(f"Int Value={self.value}")

    def gui_set_input(self) -> Optional[int]:
        imgui.set_next_item_width(100)
        changed, new_value = imgui.slider_int("", self.value, 0, 1000)
        if changed:
            return new_value
        else:
            return None

    def get(self) -> Optional[Any]:
        return self.value

    def set(self, v: Any) -> None:
        assert isinstance(v, int)
        self.value = v



from fiatlight.any_data_with_gui import AnyDataWithGui
from fiatlight.function_with_gui import FunctionWithGui
from fiatlight.functions_composition_graph import FunctionsCompositionGraph

from fiatlight.base_data_with_gui import IntWithGui
from typing import Any, Callable


PureFunction = Callable[[Any], Any]


__all__ = [
    "PureFunction",
    "AnyDataWithGui",
    "FunctionWithGui",
    "FunctionsCompositionGraph",
    "IntWithGui",
]



from typing import Any, Optional, Sequence
from imgui_bundle import imgui, imgui_node_editor as node_ed
from fiatlight import PureFunction, AnyDataWithGui, FunctionWithGui, FunctionsCompositionGraph


def versatile_gui_data(value: Any) -> None:
    if value is None:
        imgui.text("None")
    elif isinstance(value, int):
        imgui.text(f"Int Value={value}")
    elif isinstance(value, float):
        imgui.text(f"Float Value={value:.4f}")
        if imgui.is_item_hovered():
            imgui.set_tooltip(f"{value}")
    elif isinstance(value, str):
        max_len = 50
        if len(value) > max_len:
            imgui.text(f"Str len={len(value)}")
            imgui.text('"' + value[:max_len])
            if imgui.button("..."):
                imgui.open_popup("popup_value_text")

            node_ed.suspend_node_editor_canvas_immapp()
            if imgui.begin_popup("popup_value_text"):
                imgui.input_text_multiline("##value_text", value)
                imgui.end_popup()
            node_ed.resume_node_editor_canvas_immapp()
        else:
            imgui.text('"' + value + '"')
    elif isinstance(value, list):
        imgui.text(f"List len={len(value)}")
        for i, v in enumerate(value):
            if i >= 5:
                if imgui.button("..."):
                    imgui.open_popup("popup_value_list")

                # node_ed.suspend_node_editor_canvas_immapp()
                if imgui.begin_popup("popup_value_list"):
                    for i, v in enumerate(value):
                        versatile_gui_data(v)
                    imgui.end_popup()
                # node_ed.resume_node_editor_canvas_immapp()

                break
            else:
                versatile_gui_data(v)
    elif isinstance(value, tuple):
        # imgui.text(f"Tuple len={len(value)}")
        strs = [str(v) for v in value]
        tuple_str = "(" + ", ".join(strs) + ")"
        imgui.text(tuple_str)

    else:
        raise Exception(f"versatile_gui_data Unsupported type: {type(value)}")


def versatile_gui_set_input(value: Any) -> Optional[Any]:
    changed = False
    new_value: Any = None
    if value is None:
        imgui.text("gui_set_input: unknown type (None)")
    elif isinstance(value, int):
        imgui.set_next_item_width(100)
        changed, new_value = imgui.slider_int("", value, 0, 100)
    elif isinstance(value, float):
        imgui.set_next_item_width(100)
        changed, new_value = imgui.slider_float("", value, 0.0, 100.0)
    elif isinstance(value, str):
        imgui.set_next_item_width(100)
        changed, new_value = imgui.input_text("", value)
    else:
        raise Exception(f"VersatileDataWithGui Unsupported type: {type(value)}")

    if changed:
        return new_value
    else:
        return None


class VersatileDataWithGui(AnyDataWithGui):
    def gui_data(self, function_name: str) -> None:
        imgui.push_id(str(id(self)))
        versatile_gui_data(self.value)
        imgui.pop_id()

    def gui_set_input(self) -> Optional[Any]:
        return versatile_gui_set_input(self.value)


class VersatileFunctionWithGui(FunctionWithGui):
    inner_function: PureFunction
    inner_function_name: str

    def __init__(self, function: PureFunction, function_name: Optional[str] = None):
        self.inner_function = function
        if function_name is not None:
            self.inner_function_name = function_name
        else:
            self.inner_function_name = function.__name__
        self.input_gui = VersatileDataWithGui()
        self.output_gui = VersatileDataWithGui()

    def f(self, x: Any) -> Any:
        return self.inner_function(x)

    def name(self) -> str:
        return self.inner_function_name

    def gui_params(self) -> bool:
        return False


class VersatileFunctionsCompositionGraph(FunctionsCompositionGraph):
    def __init__(self, functions: Sequence[PureFunction]) -> None:
        functions_with_gui = [VersatileFunctionWithGui(f) for f in functions]
        FunctionsCompositionGraph.__init__(self, functions_with_gui)
        # super.__init__(self, functions_with_gui)



from __future__ import annotations
from typing import Any

from fiatlight.computer_vision.image_types import ImageUInt8
from fiatlight.any_data_with_gui import AnyDataWithGui
from fiatlight.computer_vision.cv_color_type import ColorType
from imgui_bundle import immvision, imgui
from imgui_bundle import portable_file_dialogs as pfd


import numpy as np
from typing import Tuple, Optional
import cv2


class ImageWithGui(AnyDataWithGui):
    # value: Optional[ImageUInt8]
    image_params: immvision.ImageParams
    open_file_dialog: Optional[pfd.open_file] = None

    def __init__(self, image: Optional[ImageUInt8] = None, zoom_key: str = "z", image_display_width: int = 200) -> None:
        self.value = image
        self.first_frame = True
        self.image_params = immvision.ImageParams()
        self.image_params.image_display_size = (image_display_width, 0)
        self.image_params.zoom_key = zoom_key

    def set(self, v: Any) -> None:
        assert v is None or type(v) == np.ndarray
        self.value = v
        self.first_frame = True

    def gui_data(self, function_name: str) -> None:
        self.image_params.refresh_image = self.first_frame
        _, self.image_params.image_display_size = gui_edit_size(self.image_params.image_display_size)
        if self.value is not None:
            immvision.image(function_name, self.value, self.image_params)
            self.first_frame = False
        if imgui.small_button("Inspect"):
            immvision.inspector_add_image(self.value, function_name)

    def gui_set_input(self) -> Optional[Any]:
        result = None
        if imgui.button("Select image file"):
            self.open_file_dialog = pfd.open_file(
                "Select image file", filters=["*.png", "*.jpg", "*.jpeg", "*.bmp", "*.tga"]
            )
        if self.open_file_dialog is not None and self.open_file_dialog.ready():
            if len(self.open_file_dialog.result()) == 1:
                image_file = self.open_file_dialog.result()[0]
                image = cv2.imread(image_file)
                if image is not None:
                    result = image
            self.open_file_dialog = None
        return result


class ImageChannelsWithGui(AnyDataWithGui):
    # value: Optional[ImageUInt8]  # We are displaying the different channels of this image
    images_params: immvision.ImageParams
    color_type: ColorType = ColorType.BGR

    def __init__(
        self,
        images: Optional[ImageUInt8] = None,  # images is a numpy of several image along the first axis
        zoom_key: str = "z",
        image_display_width: int = 200,
    ) -> None:
        self.value = images
        self.first_frame = True

        self.image_params = immvision.ImageParams()
        self.image_params.image_display_size = (image_display_width, 0)
        self.image_params.zoom_key = zoom_key

    def set(self, v: Any) -> None:
        assert type(v) == np.ndarray
        self.value = v
        self.first_frame = True

    def gui_data(self, function_name: str) -> None:
        refresh_image = self.first_frame
        self.first_frame = False

        changed, self.image_params.image_display_size = gui_edit_size(self.image_params.image_display_size)
        if self.value is not None:
            for i, image in enumerate(self.value):
                channel_name = self.color_type.channel_name(i)
                self.image_params.refresh_image = refresh_image
                label = f"{channel_name}"
                immvision.image(label, image, self.image_params)
                if imgui.small_button("Inspect"):
                    immvision.inspector_add_image(image, label)


CvSize = Tuple[int, int]


def gui_edit_size(size: CvSize) -> Tuple[bool, CvSize]:
    def modify_size_by_ratio(ratio: float) -> CvSize:
        w = int(size[0] * ratio + 0.5)
        h = int(size[1] * ratio + 0.5)
        return w, h

    changed = False
    ratio = 1.05
    imgui.push_button_repeat(True)
    imgui.text("Thumbnail size")
    imgui.same_line()
    if imgui.small_button(" smaller "):
        size = modify_size_by_ratio(1.0 / ratio)
        changed = True
    imgui.same_line()
    if imgui.small_button(" bigger "):
        size = modify_size_by_ratio(ratio)
        changed = True
    imgui.pop_button_repeat()

    return changed, size



from imgui_bundle import immapp
from fiatlight import PureFunction
from fiatlight.versatile import VersatileFunctionsCompositionGraph

import math


def main() -> None:
    functions: list[PureFunction] = [math.exp, math.sin, math.log, math.cos]
    functions_graph = VersatileFunctionsCompositionGraph(functions)

    functions_graph.set_input(2)

    def gui() -> None:
        functions_graph.draw()

    immapp.run(gui, with_node_editor=True, window_size=(1400, 600), window_title="play_versatile_math")


if __name__ == "__main__":
    main()



from __future__ import annotations
from fiatlight.any_data_with_gui import AnyDataWithGui
from fiatlight.function_with_gui import FunctionWithGui
from imgui_bundle import imgui, imgui_node_editor as ed, icons_fontawesome, ImVec2, ImVec4
from typing import List, Optional, Any, Sequence
import traceback
import sys


class FunctionsCompositionGraph:
    function_nodes: List[_FunctionNode]

    def __init__(self, functions: Sequence[FunctionWithGui]) -> None:
        assert len(functions) > 0
        f0 = functions[0]

        input_fake_function = _InputWithGui()
        input_fake_function.input_gui = f0.input_gui
        input_fake_function.output_gui = f0.input_gui

        input_node = _FunctionNode(input_fake_function)
        self.function_nodes = []
        self.function_nodes.append(input_node)

        for f in functions:
            self.function_nodes.append(_FunctionNode(f))

        for i in range(len(self.function_nodes) - 1):
            fn0 = self.function_nodes[i]
            fn1 = self.function_nodes[i + 1]
            fn0.next_function_node = fn1

    def set_input(self, input: Any) -> None:
        self.function_nodes[0].set_input(input)

    def draw(self) -> None:
        imgui.push_id(str(id(self)))

        ed.begin("FunctionsCompositionGraph")
        # draw function nodes
        for i, fn in enumerate(self.function_nodes):
            fn.draw_node(idx=i)
        # Note: those loops shall not be merged
        for fn in self.function_nodes:
            fn.draw_link()
        ed.end()

        imgui.pop_id()


class _InputWithGui(FunctionWithGui):
    def f(self, x: Any) -> Any:
        return x

    def gui_params(self) -> bool:
        return False

    def name(self) -> str:
        return "Input"


class _FunctionNode:
    function: FunctionWithGui
    next_function_node: Optional[_FunctionNode]
    input_data_with_gui: AnyDataWithGui
    output_data_with_gui: AnyDataWithGui

    last_exception_message: Optional[str] = None
    last_exception_traceback: Optional[str] = None

    node_id: ed.NodeId
    pin_input: ed.PinId
    pin_output: ed.PinId
    link_id: ed.LinkId

    def __init__(self, function: FunctionWithGui) -> None:
        self.function = function
        self.next_function_node = None
        self.input_data_with_gui = function.input_gui
        self.output_data_with_gui = function.output_gui

        self.node_id = ed.NodeId.create()
        self.pin_input = ed.PinId.create()
        self.pin_output = ed.PinId.create()
        self.link_id = ed.LinkId.create()

    def _draw_exception_message(self) -> None:
        # return
        if self.last_exception_message is None:
            return
        imgui.text_colored(ImVec4(1, 0, 0, 1), self.last_exception_message)
        # imgui.same_line()

        # popup_id = "popup_exception_details_" + self.function.name()
        # if imgui.button("..."):
        #     imgui.open_popup(popup_id)

        # if imgui.begin_popup(popup_id):
        #     #imgui.set_next_item_width(immapp.em_size(60))
        #     imgui.input_text_multiline(
        #         "##value_text",
        #                self.last_exception_traceback,
        #                 size=immapp.em_to_vec2(60, 20))
        #     imgui.end_popup()
        # # ed.resume_node_editor_canvas_immapp()

    def draw_node(self, idx: int) -> None:
        assert self.function is not None

        ed.begin_node(self.node_id)
        position = ed.get_node_position(self.node_id)
        if position.x == 0 and position.y == 0:
            width_between_nodes = 200
            position = ImVec2(idx * width_between_nodes + 1, 0)
            ed.set_node_position(self.node_id, position)

        id_fn = str(id(self.function))
        imgui.push_id(id_fn)

        imgui.text(self.function.name())
        self._draw_exception_message()

        params_changed = self.function.gui_params()
        if params_changed:
            if self.input_data_with_gui.get() is not None and self.function is not None:
                r = self.function.f(self.input_data_with_gui.get())
                self.output_data_with_gui.set(r)
                if self.next_function_node is not None:
                    self.next_function_node.set_input(r)
        imgui.pop_id()

        draw_input_pin = idx != 0
        if draw_input_pin:
            ed.begin_pin(self.pin_input, ed.PinKind.input)
            imgui.text(icons_fontawesome.ICON_FA_CIRCLE)
            ed.end_pin()

        draw_input_set_data = idx == 0
        if draw_input_set_data:
            new_value = self.input_data_with_gui.gui_set_input()
            if new_value is not None:
                self.set_input(new_value)

        def draw_output() -> None:
            if self.output_data_with_gui.get() is None:
                imgui.text("None")
            else:
                imgui.push_id(str(id(self.output_data_with_gui)))
                imgui.begin_group()
                self.output_data_with_gui.gui_data(function_name=self.function.name())
                imgui.pop_id()
                imgui.end_group()
            imgui.same_line()
            ed.begin_pin(self.pin_output, ed.PinKind.output)
            imgui.text(icons_fontawesome.ICON_FA_CIRCLE)
            ed.end_pin()

        draw_output()

        ed.end_node()

    def draw_link(self) -> None:
        if self.next_function_node is None:
            return
        ed.link(self.link_id, self.pin_output, self.next_function_node.pin_input)

    def set_input(self, input_data: Any) -> None:
        self.input_data_with_gui.set(input_data)

        if self.function is not None:
            try:
                r = self.function.f(input_data)
                self.last_exception_message = None
                self.last_exception_traceback = None
            except Exception as e:
                self.last_exception_message = str(e)
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback_details = traceback.format_exception(exc_type, exc_value, exc_traceback)
                self.last_exception_traceback = "".join(traceback_details)
                r = None

            self.output_data_with_gui.set(r)
            if self.next_function_node is not None:
                self.next_function_node.set_input(r)



from __future__ import annotations
from fiatlight.functions_composition_graph import FunctionsCompositionGraph
from fiatlight.function_with_gui import FunctionWithGui
from fiatlight.computer_vision import ImageUInt8
from fiatlight.computer_vision.image_with_gui import ImageWithGui
from fiatlight.computer_vision.lut import Split_Lut_Merge_WithGui
from fiatlight.computer_vision.cv_color_type import CvColorConversionCode, ColorType
from imgui_bundle import immapp, imgui
from typing import Any

import cv2
import numpy as np
import os


def demos_assets_folder() -> str:
    this_dir = os.path.dirname(os.path.abspath(__file__))
    assets_dir = os.path.abspath(f"{this_dir}/../../../demos_assets")
    return assets_dir


class GaussianBlurWithGui(FunctionWithGui):
    sigma_x: float = 3.0
    sigma_y: float = 3.0

    def __init__(self) -> None:
        self.input_gui = ImageWithGui()
        self.output_gui = ImageWithGui()

    def f(self, x: Any) -> ImageUInt8:
        assert type(x) == np.ndarray
        ksize = (0, 0)
        blur: ImageUInt8 = cv2.GaussianBlur(x, ksize=ksize, sigmaX=self.sigma_x, sigmaY=self.sigma_y)  # type: ignore
        return blur

    def name(self) -> str:
        return "GaussianBlur"

    def gui_params(self) -> bool:
        imgui.set_next_item_width(100)
        changed1, self.sigma_x = imgui.slider_float("sigma_x", self.sigma_x, 0.1, 15.0)
        imgui.set_next_item_width(100)
        changed2, self.sigma_y = imgui.slider_float("sigma_y", self.sigma_y, 0.1, 15.0)
        return changed1 or changed2


class CannyWithGui(FunctionWithGui):
    t_lower = 100  # Lower Threshold
    t_upper = 200  # Upper threshold
    aperture_size = 5  # Aperture size (3, 5, or 7)

    def __init__(self) -> None:
        self.input_gui = ImageWithGui()
        self.output_gui = ImageWithGui()

    def f(self, x: Any) -> ImageUInt8:
        assert type(x) == np.ndarray
        edge: ImageUInt8 = cv2.Canny(x, self.t_lower, self.t_upper, apertureSize=self.aperture_size)  # type: ignore
        return edge

    def name(self) -> str:
        return "Canny"

    def gui_params(self) -> bool:
        imgui.set_next_item_width(100)
        changed1, self.t_lower = imgui.slider_int("t_lower", self.t_lower, 0, 255)
        imgui.set_next_item_width(100)
        changed2, self.t_upper = imgui.slider_int("t_upper", self.t_upper, 0, 255)
        imgui.set_next_item_width(100)

        imgui.text("Aperture")
        imgui.same_line()
        changed3 = False
        for aperture_value in [3, 5, 7]:
            clicked: bool
            clicked, self.aperture_size = imgui.radio_button(str(aperture_value), self.aperture_size, aperture_value)
            if clicked:
                changed3 = True
            imgui.same_line()
        imgui.new_line()
        return changed1 or changed2 or changed3


class OilPaintingWithGui(FunctionWithGui):
    dynRatio = 1  # image is divided by dynRatio before histogram processing
    size = 3  # size neighbouring size is 2-size+1
    color_conversion: CvColorConversionCode  # color space conversion code

    def __init__(self) -> None:
        self.input_gui = ImageWithGui()
        self.output_gui = ImageWithGui()
        self.color_conversion = cv2.COLOR_BGR2HSV

    def f(self, x: Any) -> ImageUInt8:
        assert type(x) == np.ndarray
        r = np.zeros_like(x)
        # pip install opencv-contrib-python
        r = cv2.xphoto.oilPainting(x, self.size, self.dynRatio, self.color_conversion)  # type: ignore
        return r

    def name(self) -> str:
        return "Oil Painting"

    def gui_params(self) -> bool:
        imgui.set_next_item_width(100)
        changed1, self.dynRatio = imgui.slider_int("dynRatio", self.dynRatio, 1, 10)
        imgui.set_next_item_width(100)
        changed2, self.size = imgui.slider_int("size", self.size, 1, 10)
        return changed1 or changed2


def main() -> None:
    image = cv2.imread(demos_assets_folder() + "/images/house.jpg")
    image = cv2.resize(image, (int(image.shape[1] * 0.5), int(image.shape[0] * 0.5)))

    split_lut_merge_gui = Split_Lut_Merge_WithGui(ColorType.BGR)

    functions = [split_lut_merge_gui.split, split_lut_merge_gui.lut, split_lut_merge_gui.merge, OilPaintingWithGui()]
    # functions = [GaussianBlurWithGui(), CannyWithGui()]

    composition_graph = FunctionsCompositionGraph(functions)
    composition_graph.set_input(image)

    def gui() -> None:
        from imgui_bundle import hello_imgui

        hello_imgui.get_runner_params().fps_idling.enable_idling = False
        imgui.text(f"FPS: {imgui.get_io().framerate}")
        composition_graph.draw()

    immapp.run(gui, with_node_editor=True, window_size=(1600, 1000), fps_idle=0, window_title="fiat_image")


if __name__ == "__main__":
    main()



# # from fiatlight import FunctionsCompositionGraph
# from imgui_bundle import immapp, imgui, ImVec2, ImVec4
# from typing import Any
#
#
# class FiatLux:
#     def __init__(self) -> None:
#         self._functions_composition_graph = None
#
#     def set_functions_composition_graph(self, functions_composition_graph: Any) -> None:
#         self._functions_composition_graph = functions_composition_graph
#
#     def draw(self) -> None:
#         imgui.push_id(str(id(self)))
#         imgui.pop_id()



from typing import Any
from imgui_bundle import imgui, immapp
from fiatlight import IntWithGui, FunctionWithGui, FunctionsCompositionGraph


class AddWithGui(FunctionWithGui):
    what_to_add: int

    def __init__(self) -> None:
        self.what_to_add = 1
        self.input_gui = IntWithGui()
        self.output_gui = IntWithGui()

    def f(self, x: Any) -> int:
        assert isinstance(x, int)
        return x + self.what_to_add

    def name(self) -> str:
        return "Add"

    def gui_params(self) -> bool:
        imgui.set_next_item_width(100)
        changed, self.what_to_add = imgui.slider_int("", self.what_to_add, 0, 10)
        return changed


def main() -> None:
    functions = [AddWithGui(), AddWithGui(), AddWithGui()]
    functions_graph = FunctionsCompositionGraph(functions)

    functions_graph.set_input(1)

    def gui() -> None:
        functions_graph.draw()

    immapp.run(gui, window_size=(800, 600), window_title="fiat_add", with_node_editor=True)


if __name__ == "__main__":
    main()



from fiatlight.computer_vision.image_with_gui import ImageWithGui, ImageChannelsWithGui
from fiatlight.computer_vision.image_types import ImageUInt8, ImageFloat


__all__ = [
    "ImageUInt8",
    "ImageFloat",
    "ImageWithGui",
    "ImageChannelsWithGui",
]



from typing import List, Optional, Dict
from dataclasses import dataclass
import enum
import cv2


CvColorConversionCode = int


class ColorType(enum.Enum):
    BGR = enum.auto()
    RGB = enum.auto()
    HSV = enum.auto()
    HLS = enum.auto()
    Lab = enum.auto()
    Luv = enum.auto()
    XYZ = enum.auto()

    def channels_names(self) -> List[str]:
        return list(self.name)

    def channel_name(self, i: int) -> str:
        names = self.channels_names()
        assert 0 <= i < len(names)
        return names[i]


@dataclass
class ColorConversion:
    name: str
    src_color: ColorType
    dst_color: ColorType
    conversion_code: CvColorConversionCode


@dataclass
class ColorConversionPair:
    """Two inverse color conversions"""

    name: str
    conversion: ColorConversion
    inv_conversion: ColorConversion


def cv_color_conversion_code_between(type1: ColorType, type2: ColorType) -> Optional[CvColorConversionCode]:
    def handle_bgr() -> Optional[CvColorConversionCode]:
        if type1 == ColorType.BGR:
            conversions: Dict[ColorType, CvColorConversionCode] = {
                ColorType.RGB: cv2.COLOR_BGR2RGB,
                ColorType.HSV: cv2.COLOR_BGR2HSV_FULL,
                ColorType.HLS: cv2.COLOR_BGR2HLS_FULL,
                ColorType.Lab: cv2.COLOR_BGR2Lab,
                ColorType.Luv: cv2.COLOR_BGR2Luv,
                ColorType.XYZ: cv2.COLOR_BGR2XYZ,
            }
            if type2 in conversions.keys():
                return conversions[type2]

        if type2 == ColorType.BGR:
            conversions_inv: Dict[ColorType, CvColorConversionCode] = {
                ColorType.RGB: cv2.COLOR_RGB2BGR,
                ColorType.HSV: cv2.COLOR_HSV2BGR_FULL,
                ColorType.HLS: cv2.COLOR_HLS2BGR_FULL,
                ColorType.Lab: cv2.COLOR_Lab2BGR,
                ColorType.Luv: cv2.COLOR_Luv2BGR,
                ColorType.XYZ: cv2.COLOR_XYZ2BGR,
            }
            if type1 in conversions_inv.keys():
                return conversions_inv[type1]
        return None

    def handle_rgb() -> Optional[CvColorConversionCode]:
        if type1 == ColorType.RGB:
            conversions: Dict[ColorType, CvColorConversionCode] = {
                ColorType.BGR: cv2.COLOR_RGB2BGR,
                ColorType.HSV: cv2.COLOR_RGB2HSV_FULL,
                ColorType.HLS: cv2.COLOR_RGB2HLS_FULL,
                ColorType.Lab: cv2.COLOR_RGB2Lab,
                ColorType.Luv: cv2.COLOR_RGB2Luv,
                ColorType.XYZ: cv2.COLOR_RGB2XYZ,
            }
            if type2 in conversions.keys():
                return conversions[type2]

        if type2 == ColorType.RGB:
            conversions_inv: Dict[ColorType, CvColorConversionCode] = {
                ColorType.BGR: cv2.COLOR_BGR2RGB,
                ColorType.HSV: cv2.COLOR_HSV2RGB_FULL,
                ColorType.HLS: cv2.COLOR_HLS2RGB_FULL,
                ColorType.Lab: cv2.COLOR_Lab2RGB,
                ColorType.Luv: cv2.COLOR_Luv2RGB,
                ColorType.XYZ: cv2.COLOR_XYZ2RGB,
            }
            if type1 in conversions_inv.keys():
                return conversions_inv[type1]

        return None

    with_bgr = handle_bgr()
    if with_bgr is not None:
        return with_bgr

    with_rgb = handle_rgb()
    if with_rgb is not None:
        return with_rgb

    return None


def compute_possible_conversion_pairs(color_type: ColorType) -> List[ColorConversionPair]:
    r: List[ColorConversionPair] = []
    for other_color_type in ColorType:
        conversion_code = cv_color_conversion_code_between(color_type, other_color_type)
        conversion_code_inv = cv_color_conversion_code_between(other_color_type, color_type)
        if conversion_code is not None and conversion_code_inv is not None:
            conversion_direct = ColorConversion(
                f"{color_type.name}=>{other_color_type.name}", color_type, other_color_type, conversion_code
            )
            conversion_inv = ColorConversion(
                f"{other_color_type.name}=>{color_type.name}", other_color_type, color_type, conversion_code_inv
            )
            conversion_pair = ColorConversionPair(
                f"{color_type.name}=>{other_color_type.name}=>{color_type.name}", conversion_direct, conversion_inv
            )
            r.append(conversion_pair)
    return r



from typing import Any
from abc import ABC, abstractmethod
from fiatlight.any_data_with_gui import AnyDataWithGui


class FunctionWithGui(ABC):
    """Override this class with your functions which you want to vizualize in a graph
    // FunctionWithGui: any function that can be presented visually, with
    // - a displayed name
    // - a gui in order to modify the internal params
    // - a pure function f: AnyDataWithGui -> AnyDataWithGui
    """

    # input_gui and output_gui should be filled during construction
    input_gui: AnyDataWithGui
    output_gui: AnyDataWithGui

    @abstractmethod
    def f(self, x: Any) -> Any:
        pass

    @abstractmethod
    def name(self) -> str:
        pass

    def gui_params(self) -> bool:
        """override this if you want to provide a gui for the function inner params
        (i.e. neither input nor output params, but the function internal state)
        It should return True if the inner params were changed.
        """
        return False



from typing import Optional, Any
from abc import ABC, abstractmethod


class AnyDataWithGui(ABC):
    """
    Override this class with your types, and implement a draw function that presents it content
    """

    value: Any

    @abstractmethod
    def gui_data(self, function_name: str) -> None:
        """Override this by implementing a draw function that presents the data content"""
        pass

    def gui_set_input(self) -> Optional[Any]:
        """Override this if you want to provide a visual way to set the input of
        a function composition graph"""
        return None

    def set(self, v: Any) -> None:
        """Override this if you want to add more behavior when setting the value"""
        self.value = v

    def get(self) -> Optional[Any]:
        return self.value



from typing import Tuple, Optional, cast, Any, List
from numpy.typing import NDArray
import cv2
import numpy as np
import math
from fiatlight import FunctionWithGui
from fiatlight.computer_vision import ImageUInt8, ImageFloat, ImageWithGui, ImageChannelsWithGui
from fiatlight.computer_vision.cv_color_type import ColorType, ColorConversionPair, compute_possible_conversion_pairs
from fiatlight.computer_vision.split_merge import SplitChannelsWithGui, MergeChannelsWithGui
from imgui_bundle import immapp, imgui, immvision


Point2d = Tuple[float, float]


class LutImage:
    pow_exponent: float = 1.0
    min_in: float = 0.0
    min_out: float = 0.0
    max_in: float = 1.0
    max_out: float = 1.0

    _lut_table: NDArray[np.float_]
    _lut_graph: ImageUInt8
    _lut_graph_needs_refresh: bool = True

    def apply(self, image: ImageFloat) -> ImageFloat:
        if not hasattr(self, "_lut_table"):
            self._prepare_lut()
        lut_uint8 = (self._lut_table * 255.0).astype(np.uint8)
        image_uint8: ImageUInt8 = (image * 255.0).astype(np.uint8)
        image_with_lut_uint8 = np.zeros_like(image_uint8)
        cv2.LUT(image_uint8, lut_uint8, image_with_lut_uint8)
        image_adjusted: ImageFloat = image_with_lut_uint8.astype(float) / 255.0
        return image_adjusted

    @staticmethod
    def _lut_graph_size() -> float:
        graph_size = int(immapp.em_size() * 2.0)
        return graph_size

    def _show_lut_graph(self, channel_name: str) -> Point2d:
        if not hasattr(self, "_lut_graph"):
            self._prepare_lut_graph()
        mouse_position = immvision.image_display(
            channel_name, self._lut_graph, refresh_image=self._lut_graph_needs_refresh
        )
        self._lut_graph_needs_refresh = False
        return mouse_position

    def _prepare_lut_graph(self) -> None:
        graph_size = self._lut_graph_size()
        x = np.arange(0.0, 1.0, 1.0 / 256.0)
        self._lut_graph = immvision._draw_lut_graph(list(x), list(self._lut_table), (graph_size, graph_size))  # type: ignore
        self._lut_graph_needs_refresh = True

    def _prepare_lut(self) -> None:
        x = np.arange(0.0, 1.0, 1.0 / 256.0)
        y = (x - self.min_in) / (self.max_in - self.min_in)
        y = np.clip(y, 0.0, 1.0)
        y = np.power(y, self.pow_exponent)
        y = np.clip(y, 0.0, 1.0)
        y = self.min_out + (self.max_out - self.min_out) * y
        y = np.clip(y, 0.0, 1.0)
        self._lut_table = y

    def handle_graph_mouse_edit(self, mouse_position: Point2d) -> bool:
        drag_threshold = 0
        mouse_button = 0
        changed = False

        def get_mouse_position_normalized() -> Optional[Point2d]:
            r: Optional[Point2d]
            graph_size = self._lut_graph_size()
            if mouse_position[0] >= 0:
                r = (mouse_position[0] / graph_size, 1 - mouse_position[1] / graph_size)
                return r
            else:
                return None

        mouse_position_normalized = get_mouse_position_normalized()

        if mouse_position_normalized is not None:
            imgui.text(f"{mouse_position_normalized[0]:.2f}, {mouse_position_normalized[1]:.2f}")

        if mouse_position_normalized is not None and imgui.is_mouse_dragging(0, drag_threshold):
            drag_delta = imgui.get_mouse_drag_delta(mouse_button)
            drag_horizontal = math.fabs(drag_delta.x) > math.fabs(drag_delta.y)
            drag_vertical = not drag_horizontal
            imgui.reset_mouse_drag_delta(mouse_button)

            delta_edge = 0.37
            if drag_horizontal and mouse_position_normalized[1] < delta_edge:
                self.min_in = mouse_position_normalized[0]
                changed = True
            elif drag_horizontal and mouse_position_normalized[1] > 1.0 - delta_edge:
                self.max_in = mouse_position_normalized[0]
                changed = True
            elif drag_vertical and mouse_position_normalized[0] < delta_edge:
                self.min_out = mouse_position_normalized[1]
                changed = True
            elif drag_vertical and mouse_position_normalized[0] > 1.0 - delta_edge:
                self.max_out = mouse_position_normalized[1]
                changed = True

        return changed

    def gui_params(self, channel_name: str) -> bool:
        changed = False

        imgui.begin_group()
        mouse_position = self._show_lut_graph(channel_name)
        if self.handle_graph_mouse_edit(mouse_position):
            changed = True

        if imgui.small_button("Reset"):
            self.min_in, self.max_in = (0.0, 1.0)
            self.min_out, self.max_out = (0.0, 1.0)
            self.pow_exponent = 1.0
            changed = True
        imgui.end_group()

        imgui.same_line()

        imgui.begin_group()
        idx_slider = 0

        def show_slider(label: str, v: float, min: float, max: float, logarithmic: bool) -> float:
            nonlocal idx_slider, changed
            imgui.set_next_item_width(70)
            idx_slider += 1
            flags = imgui.SliderFlags_.logarithmic.value if logarithmic else 0
            edited_this_slider, v = imgui.slider_float(f"{label}##slider{idx_slider}", v, min, max, flags=flags)
            if edited_this_slider:
                changed = True
            return v

        def show_01_slider(label: str, v: float) -> float:
            return show_slider(label, v, 0, 1, False)

        def show_two_01_sliders(label: str, v_min: float, v_max: float) -> Tuple[float, float]:
            v_min = show_01_slider(f"##{label}v_min", v_min)
            imgui.same_line()
            v_max = show_01_slider(f"{label}##v_max", v_max)
            if math.fabs(v_max - v_min) < 1e-3:  # avoid div by 0
                v_min = v_max - 0.01
            return v_min, v_max

        self.pow_exponent = show_slider("Gamma power", self.pow_exponent, 0.0, 10.0, True)
        self.min_in, self.max_in = show_two_01_sliders("In", self.min_in, self.max_in)
        self.min_out, self.max_out = show_two_01_sliders("Out", self.min_out, self.max_out)

        if changed:
            self._prepare_lut()
            self._prepare_lut_graph()
        imgui.end_group()
        return changed


class LutImageWithGui(FunctionWithGui):
    lut_image: LutImage

    def __init__(self) -> None:
        self.lut_image = LutImage()
        self.input_gui = ImageWithGui()
        self.output_gui = ImageWithGui()

    def f(self, x: Any) -> Any:
        assert type(x) == ImageFloat
        image_adjusted = self.lut_image.apply(x)
        return image_adjusted

    def name(self) -> str:
        return "LUT"

    def gui_params(self) -> bool:
        return self.lut_image.gui_params("LUT")


class LutChannelsWithGui(FunctionWithGui):
    color_type: ColorType = ColorType.BGR
    channel_adjust_params: List[LutImage]

    def __init__(self) -> None:
        self.input_gui = ImageChannelsWithGui()
        self.output_gui = ImageChannelsWithGui()

    def output_gui_channels(self) -> ImageChannelsWithGui:
        return cast(ImageChannelsWithGui, self.output_gui)

    def add_params_on_demand(self, nb_channels: int) -> None:
        if not hasattr(self, "channel_adjust_params"):
            self.channel_adjust_params = []
        while len(self.channel_adjust_params) < nb_channels:
            self.channel_adjust_params.append(LutImage())

    def f(self, x: Any) -> Any:
        assert type(x) == np.ndarray

        original_channels = x
        self.add_params_on_demand(len(original_channels))

        adjusted_channels = np.zeros_like(original_channels)
        for i in range(len(original_channels)):
            adjusted_channels[i] = self.channel_adjust_params[i].apply(original_channels[i])

        return adjusted_channels

    def name(self) -> str:
        return "LUT channels"

    def gui_params(self) -> bool:
        changed = False
        for i, channel_adjust_param in enumerate(self.channel_adjust_params):
            channel_name = self.color_type.channels_names()[i]
            imgui.push_id(str(i))
            changed |= channel_adjust_param.gui_params(channel_name)
            imgui.pop_id()
        return changed


class Split_Lut_Merge_WithGui:
    possible_conversion_pairs: List[ColorConversionPair]
    current_conversion_pair: Optional[ColorConversionPair]
    show_possible_color_conversions: bool = False
    split: SplitChannelsWithGui
    lut: LutChannelsWithGui
    merge: MergeChannelsWithGui

    def __init__(self, color_type: ColorType):
        self.possible_conversion_pairs = compute_possible_conversion_pairs(color_type)
        self.current_conversion_pair = None
        self.split = SplitChannelsWithGui()
        self.split.gui_params_optional_fn = lambda: self.gui_select_conversion()
        self.merge = MergeChannelsWithGui()
        self.lut = LutChannelsWithGui()

    def gui_select_conversion(self) -> bool:
        changed = False
        _, self.show_possible_color_conversions = imgui.checkbox(
            "Show Color Conversions", self.show_possible_color_conversions
        )
        if self.show_possible_color_conversions:
            if imgui.radio_button("None", self.current_conversion_pair is None):
                changed = True
                self.current_conversion_pair = None
            for conversion_pair in self.possible_conversion_pairs:
                active = self.current_conversion_pair == conversion_pair
                if imgui.radio_button(conversion_pair.conversion.name, active):
                    self.current_conversion_pair = conversion_pair
                    changed = True

        if changed:
            if self.current_conversion_pair is None:
                self.split.color_conversion = None
                self.merge.color_conversion = None
                # self.lut.output_gui.color_type = ColorType.BGR
            else:
                self.split.color_conversion = self.current_conversion_pair.conversion
                self.merge.color_conversion = self.current_conversion_pair.inv_conversion

                self.lut.color_type = self.current_conversion_pair.conversion.dst_color
                self.lut.output_gui_channels().color_type = self.current_conversion_pair.conversion.dst_color
                self.split.output_gui_channels().color_type = self.current_conversion_pair.conversion.dst_color
                print("a")
        return changed
